use std::time::Duration as StdDuration;

use serde::{Serialize, Deserialize};
use chrono::naive::serde::ts_seconds::serialize as ts_seconds_naive;
use chrono::{Duration as ChronoDuration, NaiveDateTime, Utc, DateTime};
use diesel::prelude::*;
use rand::Rng;
use rand::rngs::OsRng;
use rand::distributions::Alphanumeric;

use crate::models::user::UserInfo;
use crate::schema::*;
use crate::models::errors::UserError;


#[derive(Debug, Deserialize)]
pub struct AuthTokenRequest {
    pub username: String,
    pub password: String,
}

#[derive(Debug, Serialize, Queryable, Identifiable, Insertable)]
#[table_name = "session"]
#[primary_key(session_id)]
pub struct Session {
    #[serde(rename = "token")]
    pub session_id: String,
    #[serde(skip)]
    pub user_id: String,
    #[serde(serialize_with = "ts_seconds_naive")]
    pub expires_at: NaiveDateTime,
}

impl Session {
    pub fn generate_id() -> String {
        OsRng
            .sample_iter(&Alphanumeric)
            .take(50)
            .map(char::from)
            .collect()
    }

    pub fn from_session_id(conn: &diesel::SqliteConnection, id: &str) -> Result<Session, UserError> {
        use crate::schema::session::dsl::*;
        session
            .find(id)
            .get_result(conn)
            .map_err(|_| UserError::ExpiredSession)
            .and_then(|s: Session| {
                let expires = DateTime::<Utc>::from_utc(s.expires_at, Utc);
                if expires < Utc::now() {
                    Err(UserError::ExpiredSession)
                } else {
                    Ok(s)
                }
            })
    }

    pub fn new(conn: &diesel::SqliteConnection, user_info: &UserInfo, token_duration: StdDuration) -> Result<Session, UserError> {
        use crate::schema::session::dsl::*;

        let expires = Utc::now() + ChronoDuration::from_std(token_duration).unwrap();

        let user_session = Session {
            session_id: Session::generate_id(),
            user_id: user_info.id.clone(),
            expires_at: expires.naive_utc(),
        };

        diesel::insert_into(session)
                .values(&user_session)
                .execute(conn)
                .map_err(UserError::DbError)?;

        Ok(user_session)

    }
}
